use chrono::{DateTime, Utc};
use sqlx::{
    postgres::{PgConnectOptions, PgPoolOptions, PgQueryResult, PgRow},
    Executor, Pool, Postgres, Row,
};

use crate::template::Post;

pub struct Db {
    conn: Pool<Postgres>,
}

impl Db {
    pub async fn new(path: &str) -> anyhow::Result<Self> {
        let conn_options: PgConnectOptions = path.parse()?;
        let conn = PgPoolOptions::new().connect_with(conn_options).await?;
        conn.execute(include_str!("sql/schema.sql")).await?;
        Ok(Db { conn })
    }

    pub async fn get_posts(&self) -> anyhow::Result<Vec<Post>> {
        let posts = sqlx::query("SELECT (id, post, ip, show_ip, image, time) FROM posts ORDER BY id")
            .map(|v: PgRow| v.get::<(i32, String, String, bool, Option<String>, DateTime<Utc>), _>(0))
            .fetch_all(&self.conn)
            .await?;

        let mut builder = Vec::new();
        for post in posts {
            let date = post.5.format("%Y-%m-%d %H:%m").to_string();
            builder.push(Post {
                id: post.0,
                content: post.1,
                ip: post.2,
                show_ip: post.3,
                image: post.4,
                time: post.5,
                date,
            });
        }
        builder.reverse();
        Ok(builder)
    }

    pub async fn add_post(
        &self,
        post: &str,
        ip: &str,
        show_ip: bool,
        image: Option<String>,
    ) -> sqlx::Result<PgQueryResult> {
        let now = chrono::offset::Utc::now();
        if let Some(image) = image {
            sqlx::query("INSERT INTO posts (post, ip, show_ip, image, time) VALUES ($1, $2, $3, $4, $5)")
                .bind(post)
                .bind(ip)
                .bind(show_ip)
                .bind(image)
                .bind(now)
                .execute(&self.conn)
                .await
        } else {
            sqlx::query("INSERT INTO posts (post, ip, show_ip, time) VALUES ($1, $2, $3, $4)")
                .bind(post)
                .bind(ip)
                .bind(show_ip)
                .bind(now)
                .execute(&self.conn)
                .await
        }
    }
}
