use crate::{db::Db, template};
use actix_web::{
    dev::Server, http::StatusCode, web, HttpRequest, HttpResponse, HttpResponseBuilder, HttpServer,
    Responder,
};
use askama::Template;
use futures_util::{stream::StreamExt, TryStreamExt};
use std::{
    collections::hash_map::RandomState,
    hash::{BuildHasher, Hash, Hasher},
    time,
};
use std::{net::SocketAddr, sync::Arc};
use tokio::fs;
use askama_escape::Escaper;

pub struct WebServer {
    pub server: Server,
}

impl WebServer {
    pub fn new(addr: SocketAddr, db: Db, image_path: String) -> anyhow::Result<Self> {
        let db = Arc::new(db);
        let server = HttpServer::new(move || {
            actix_web::App::new()
                .app_data(web::Data::new(db.clone()))
                .app_data(web::Data::new(image_path.clone()))
                .route("/static/{file_name}", web::get().to(files))
                .route("/assets/{file_name}", web::get().to(assets))
                .route("/", web::get().to(index))
                .route("/add", web::get().to(add_get))
                .route("/add", web::post().to(add_post))
                .route("/rss.xml", web::get().to(rss))
        })
        .bind(addr)?
        .run();
        Ok(WebServer { server })
    }
}

async fn index(db: web::Data<Arc<Db>>) -> impl Responder {
    let posts = db.get_posts().await;
    if let Err(_) = posts {
        return HttpResponse::InternalServerError().finish();
    }
    let posts = posts.unwrap();
    let index = template::Index { posts };
    let html = index.render().unwrap();
    HttpResponseBuilder::new(StatusCode::OK)
        .content_type("text/html")
        .body(html)
}

async fn add_get() -> impl Responder {
    let add = template::Add {};
    let html = add.render().unwrap();
    HttpResponseBuilder::new(StatusCode::OK)
        .content_type("text/html")
        .body(html)
}

async fn add_post(
    mut post: actix_multipart::Multipart,
    http_req: HttpRequest,
    db: web::Data<Arc<Db>>,
    dir: web::Data<String>,
) -> Result<HttpResponse, actix_web::Error> {
    let ip = http_req
        .connection_info()
        .realip_remote_addr()
        .unwrap_or("0.0.0.0")
        .to_string();

    let mut path = None;
    let mut req = template::AddPost {
        post: None,
        show_ip: false,
        filename: None,
        image: Vec::new(),
    };

    while let Some(mut field) = post.try_next().await? {
        match field.name() {
            "post" => {
                let mut post_buf = Vec::new();
                while let Some(post) = field.next().await.transpose().unwrap() {
                    post_buf.append(&mut post.to_vec());
                }
                if !post_buf.is_empty() {
                    req.post = Some(std::str::from_utf8(&post_buf)?.to_string());   
                }
            }
            "show_ip" => req.show_ip = true,
            "image" => {
                let rand = RandomState::new();
                let mut hasher = rand.build_hasher();
                time::SystemTime::now().hash(&mut hasher);
                let hash = hasher.finish();

                req.filename = Some(format!(
                    "{hash:x}_{}",
                    field
                        .content_disposition()
                        .get_filename()
                        .unwrap_or("image")
                ));

                while let Some(image) = field.next().await.transpose()? {
                    req.image.append(&mut image.to_vec());
                }

                let file = file_format::FileFormat::from_bytes(&req.image);
                if file.media_type() != field.content_type().unwrap().essence_str() {
                    return Ok(text_plain_error(
                        "The mime type and content of the file don't match up",
                    ));
                }
            }
            _ => return Ok(text_plain_error("Received invalid form data")),
        }
    }

    if req.post.is_none() {
        return Ok(text_plain_error("Post filed cannot be empty"));
    }

    if req.image.len() > 0 {
        path = Some(format!("/static/{}", req.filename.clone().unwrap()));
        let fs_path = format!("{}/{}", dir.to_string(), req.filename.unwrap());
        fs::write(fs_path, req.image).await?;
    }

    if let Err(_) = db
        .add_post(&req.post.unwrap(), &ip, req.show_ip, path)
        .await
    {
        Ok(HttpResponse::InternalServerError().finish())
    } else {
        Ok(HttpResponse::SeeOther()
            .insert_header(("Location", "/"))
            .finish())
    }
}

async fn files(path: web::Path<String>, base_path: web::Data<String>) -> impl Responder {
    let path = format!("{}/{path}", base_path.into_inner());
    actix_files::NamedFile::open(path)
}

async fn assets(path: web::Path<String>) -> impl Responder {
    let path = format!("static/{path}");
    actix_files::NamedFile::open(path)
}

async fn rss(db: web::Data<Arc<Db>>) -> impl Responder {
    let posts = db.get_posts().await;
    if let Err(_) = posts {
        return HttpResponse::InternalServerError().finish();
    }

    let posts = posts.unwrap();
    let mut rss_builder = String::new();
    rss_builder.push_str("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>");
    rss_builder.push_str("<rss version=\"2.0\">");
    rss_builder.push_str("<channel>");
    rss_builder.push_str("<title>urandom.catgirls.dn42</title>");
    rss_builder.push_str("<link>https://urandom.catgirls.dn42</link>");
    rss_builder.push_str("<description>The home of dn42 shitposts</description>");
    for post in posts {
        rss_builder.push_str("<item>");
        let pub_date= format!("<pubDate>{}</pubDate>", post.time.format("R"));
        rss_builder.push_str(&pub_date);
        let title = if post.show_ip {
            format!("<title>Post {} - {}</title>", post.id, post.ip)
        } else {
            format!("<title>Post {}</title>", post.id)
        };
        rss_builder.push_str(&title);
        let mut post_content = String::new();
        askama_escape::Html.write_escaped(&mut post_content, &post.content).unwrap();
        let description = if let Some(image) = post.image {
            format!("<description><![CDATA[{}<br><img src=\"{image}\" alt=\"Attached image: {image}\"/>]]></description>", post_content)
        } else {
            format!("<description>{}</description>", post_content)
        };
        rss_builder.push_str(&description);
        rss_builder.push_str("</item>");
    }
    rss_builder.push_str("</channel>");
    rss_builder.push_str("</rss>");

    HttpResponseBuilder::new(StatusCode::OK)
        .content_type("application/rss+xml")
        .body(rss_builder)
}

fn text_plain_error(err: &'static str) -> HttpResponse {
    HttpResponseBuilder::new(StatusCode::OK)
        .content_type("text/plain")
        .body(err)
}
